@@ -7,6 +7,7 @@
 #pragma once
 
 #include <cstdint>
+#include <cstring>
 #include <exception>
 #include <memory>
 #include <string>
@@ -16,14 +17,7 @@
 
 namespace bioparser {
 
-static const std::string version = "v2.0.1";
-
-constexpr std::uint32_t kBufferSize = 64 * 1024;
-
-// Small/Medium/Large Storage Size
-constexpr std::uint32_t kSSS = 4 * 1024;
-constexpr std::uint32_t kMSS = 8 * 1024 * 1024;
-constexpr std::uint32_t kLSS = 512 * 1024 * 1024;
+static const std::string version = "v2.1.2";
 
 /*!
  * @brief Parser absctract class
@@ -67,7 +61,6 @@ public:
 
     bool parse(std::vector<std::shared_ptr<T>>& dst, std::uint64_t max_bytes,
         bool trim = true);
-
 protected:
     Parser(gzFile input_file, std::uint32_t storage_size);
     Parser(const Parser&) = delete;
@@ -75,6 +68,8 @@ protected:
 
     std::unique_ptr<gzFile_s, int(*)(gzFile)> input_file_;
     std::vector<char> buffer_;
+    std::uint32_t buffer_ptr_;
+    std::uint32_t buffer_bytes_;
     std::vector<char> storage_;
 };
 
@@ -88,7 +83,6 @@ public:
 
     friend std::unique_ptr<Parser<T>>
         createParser<bioparser::FastaParser, T>(const std::string& path);
-
 private:
     FastaParser(gzFile input_file);
     FastaParser(const FastaParser&) = delete;
@@ -105,7 +99,6 @@ public:
 
     friend std::unique_ptr<Parser<T>>
         createParser<bioparser::FastqParser, T>(const std::string& path);
-
 private:
     FastqParser(gzFile input_file);
     FastqParser(const FastqParser&) = delete;
@@ -122,7 +115,6 @@ public:
 
     friend std::unique_ptr<Parser<T>>
         createParser<bioparser::MhapParser, T>(const std::string& path);
-
 private:
     MhapParser(gzFile input_file);
     MhapParser(const MhapParser&) = delete;
@@ -139,7 +131,6 @@ public:
 
     friend std::unique_ptr<Parser<T>>
         createParser<bioparser::PafParser, T>(const std::string& path);
-
 private:
     PafParser(gzFile input_file);
     PafParser(const PafParser&) = delete;
@@ -156,7 +147,6 @@ public:
 
     friend std::unique_ptr<Parser<T>>
         createParser<bioparser::SamParser, T>(const std::string& path);
-
 private:
     SamParser(gzFile input_file);
     SamParser(const SamParser&) = delete;
@@ -195,8 +185,8 @@ inline std::unique_ptr<Parser<T>> createParser(const std::string& path) {
 
 template<class T>
 inline Parser<T>::Parser(gzFile input_file, std::uint32_t storage_size)
-        : input_file_(input_file, gzclose), buffer_(kBufferSize, 0),
-        storage_(storage_size, 0) {
+        : input_file_(input_file, gzclose), buffer_(65536, 0),
+        buffer_ptr_(0), buffer_bytes_(0), storage_(storage_size, 0) {
 }
 
 template<class T>
@@ -205,7 +195,9 @@ inline Parser<T>::~Parser() {
 
 template<class T>
 inline void Parser<T>::reset() {
-    gzseek(this->input_file_.get(), 0, SEEK_SET);
+    gzseek(input_file_.get(), 0, SEEK_SET);
+    buffer_ptr_ = 0;
+    buffer_bytes_ = 0;
 }
 
 template<class T>
@@ -213,7 +205,7 @@ inline bool Parser<T>::parse(std::vector<std::shared_ptr<T>>& dst,
     std::uint64_t max_bytes, bool trim) {
 
     std::vector<std::unique_ptr<T>> tmp;
-    auto ret = this->parse_objects(tmp, max_bytes, trim);
+    auto ret = parse_objects(tmp, max_bytes, trim);
 
     dst.reserve(dst.size() + tmp.size());
     for (auto& it: tmp) {
@@ -224,7 +216,7 @@ inline bool Parser<T>::parse(std::vector<std::shared_ptr<T>>& dst,
 
 template<class T>
 inline FastaParser<T>::FastaParser(gzFile input_file)
-        : Parser<T>(input_file, kSSS + kMSS) {
+        : Parser<T>(input_file, 4194304) {
 }
 
 template<class T>
@@ -236,108 +228,97 @@ inline bool FastaParser<T>::parse(std::vector<std::unique_ptr<T>>& dst,
     std::uint64_t max_bytes, bool trim) {
 
     auto input_file = this->input_file_.get();
-    bool is_end = gzeof(input_file);
-    bool is_valid = false;
-    bool status = false;
-    std::uint64_t current_bytes = 0;
+    bool is_eof = false;
+
     std::uint64_t total_bytes = 0;
-    std::uint64_t num_objects = 0;
-    std::uint64_t last_object_id = num_objects;
-    std::uint32_t line_number = 0;
-
-    char* name = &(this->storage_[0]);
-    std::uint32_t name_length = 0;
-
-    char* sequence = &(this->storage_[kSSS]);
-    std::uint32_t sequence_length = 0;
+    std::uint32_t line_num = 0;
+    std::uint32_t storage_ptr = 0;
+    std::uint32_t data_ptr = 0;
 
     auto create_T = [&] () -> void {
-        if (trim) {
-            rightStripHard(name, name_length);
-        } else {
-            rightStrip(name, name_length);
-        }
-        rightStrip(sequence, sequence_length);
 
-        if (name_length == 0 || name[0] != '>' || sequence_length == 0) {
+        if (data_ptr == 0) {
+            throw std::invalid_argument("[bioparser::FastaParser] error: "
+                "invalid file format!");
+        }
+
+        std::uint32_t name_len = data_ptr;
+        if (trim) {
+            rightStripHard(&(this->storage_[0]), name_len);
+        } else {
+            rightStrip(&(this->storage_[0]), name_len);
+        }
+        std::uint32_t data_len = storage_ptr - data_ptr;
+        rightStrip(&(this->storage_[data_ptr]), data_len);
+
+        if (name_len == 0 || this->storage_[0] != '>' || data_len == 0) {
             throw std::invalid_argument("[bioparser::FastaParser] error: "
                 "invalid file format!");
         }
 
         dst.emplace_back(std::unique_ptr<T>(new T(
-            (const char*) &(name[1]), name_length - 1,
-            (const char*) sequence, sequence_length)));
+            (const char*) &(this->storage_[1]), name_len - 1,
+            (const char*) &(this->storage_[data_ptr]), data_len)));
 
-        ++num_objects;
-        current_bytes = 1;
-        name_length = 1;
-        sequence_length = 0;
-        is_valid = false;
+        total_bytes += storage_ptr;
+        storage_ptr = 0;
+        data_ptr = 0;
     };
 
-    while (!is_end) {
+    while (true) {
 
-        std::uint64_t read_bytes = gzfread(this->buffer_.data(), sizeof(char),
-            this->buffer_.size(), input_file);
-        is_end = gzeof(input_file);
-
-        total_bytes += read_bytes;
-        if (max_bytes != 0 && total_bytes > max_bytes) {
-            if (last_object_id == num_objects) {
-                throw std::invalid_argument("[bioparser::FastaParser] error: "
-                    "too small chunk size!");
+        std::uint32_t begin_ptr = this->buffer_ptr_;
+        for (; this->buffer_ptr_ < this->buffer_bytes_; ++this->buffer_ptr_) {
+            auto c = this->buffer_[this->buffer_ptr_];
+            if (c == '\n') {
+                std::memcpy(&this->storage_[storage_ptr],
+                    &this->buffer_[begin_ptr],
+                    this->buffer_ptr_ - begin_ptr);
+                storage_ptr += this->buffer_ptr_ - begin_ptr;
+                begin_ptr = this->buffer_ptr_ + 1;
+                if (line_num == 0) {
+                    data_ptr = storage_ptr;
+                    line_num = 1;
+                }
+            } else if (line_num == 1 && c == '>') {
+                line_num = 0;
+                create_T();
+                if (total_bytes >= max_bytes) {
+                    return true;
+                }
             }
-            gzseek(input_file, -(current_bytes + read_bytes), SEEK_CUR);
-            status = true;
+        }
+        if (begin_ptr < this->buffer_ptr_) {
+            std::memcpy(&this->storage_[storage_ptr],
+                &this->buffer_[begin_ptr],
+                this->buffer_ptr_ - begin_ptr);
+            storage_ptr += this->buffer_ptr_ - begin_ptr;
+        }
+        this->buffer_ptr_ = 0;
+
+        if (is_eof) {
             break;
         }
 
-        for (std::uint32_t i = 0; i < read_bytes; ++i) {
-            auto c = this->buffer_[i];
+        this->buffer_bytes_ = gzread(input_file, this->buffer_.data(),
+            this->buffer_.size());
+        is_eof = this->buffer_bytes_ < this->buffer_.size();
 
-            if (c == '\n') {
-                ++line_number;
-            } else if (c == '>' && line_number != 0) {
-                is_valid = true;
-                line_number = 0;
-            } else {
-                switch (line_number) {
-                    case 0:
-                        if (name_length < kSSS) {
-                            if (!(name_length == 0 && isspace(c))) {
-                                name[name_length++] = c;
-                            }
-                        }
-                        break;
-                    default:
-                        sequence[sequence_length++] = c;
-                        if (sequence_length == kMSS) {
-                            this->storage_.resize(kSSS + kLSS, 0);
-                            name = &(this->storage_[0]);
-                            sequence = &(this->storage_[kSSS]);
-                        }
-                        break;
-                }
-            }
-
-            ++current_bytes;
-
-            if (is_valid) {
-                create_T();
-            }
-        }
-
-        if (is_end && current_bytes != 0) {
-            create_T();
+        if (storage_ptr + this->buffer_bytes_ > this->storage_.size()) {
+            this->storage_.resize(this->storage_.size() * 2);
         }
     }
 
-    return status;
+    if (storage_ptr != 0) {
+        create_T();
+    }
+
+    return false;
 }
 
 template<class T>
 inline FastqParser<T>::FastqParser(gzFile input_file)
-        : Parser<T>(input_file, kSSS + 2 * kMSS) {
+        : Parser<T>(input_file, 4194304) {
 }
 
 template<class T>
@@ -349,130 +330,112 @@ inline bool FastqParser<T>::parse(std::vector<std::unique_ptr<T>>& dst,
     std::uint64_t max_bytes, bool trim) {
 
     auto input_file = this->input_file_.get();
-    bool is_end = gzeof(input_file);
-    bool is_valid = false;
-    bool status = false;
-    std::uint64_t current_bytes = 0;
+    bool is_eof = false;
+
     std::uint64_t total_bytes = 0;
-    std::uint64_t num_objects = 0;
-    std::uint64_t last_object_id = num_objects;
-    std::uint32_t line_number = 0;
-
-    char* name = &(this->storage_[0]);
-    std::uint32_t name_length = 0;
-
-    char* sequence = &(this->storage_[kSSS]);
-    std::uint32_t sequence_length = 0;
-
-    char* quality = &(this->storage_[kSSS + kMSS]);
-    std::uint32_t quality_length = 0;
+    std::uint32_t line_num = 0;
+    std::uint32_t storage_ptr = 0;
+    std::uint32_t data_ptr = 0;
+    std::uint32_t comm_ptr = 0;
+    std::uint32_t qual_ptr = 0;
 
     auto create_T = [&] () -> void {
-        if (trim) {
-            rightStripHard(name, name_length);
-        } else {
-            rightStrip(name, name_length);
-        }
-        rightStrip(sequence, sequence_length);
-        rightStrip(quality, quality_length);
 
-        if (name_length == 0 || name[0] != '@' || sequence_length == 0 ||
-            quality_length == 0 || sequence_length != quality_length) {
+        if (data_ptr == 0 || comm_ptr == 0 || qual_ptr == 0) {
+            throw std::invalid_argument("[bioparser::FastqParser] error: "
+                "invalid file format!");
+        }
+
+        std::uint32_t name_len = data_ptr;
+        if (trim) {
+            rightStripHard(&(this->storage_[0]), name_len);
+        } else {
+            rightStrip(&(this->storage_[0]), name_len);
+        }
+        std::uint32_t data_len = comm_ptr - data_ptr;
+        rightStrip(&(this->storage_[data_ptr]), data_len);
+        std::uint32_t qual_len = storage_ptr - qual_ptr;
+        rightStrip(&(this->storage_[qual_ptr]), qual_len);
+
+        if (name_len == 0 || this->storage_[0] != '@' || data_len == 0 ||
+            qual_len == 0 || data_len != qual_len) {
             throw std::invalid_argument("[bioparser::FastqParser] error: "
                 "invalid file format!");
         }
 
         dst.emplace_back(std::unique_ptr<T>(new T(
-            (const char*) &(name[1]), name_length - 1,
-            (const char*) sequence, sequence_length,
-            (const char*) quality, quality_length)));
+            (const char*) &(this->storage_[1]), name_len - 1,
+            (const char*) &(this->storage_[data_ptr]), data_len,
+            (const char*) &(this->storage_[qual_ptr]), qual_len)));
 
-        ++num_objects;
-        current_bytes = 0;
-        name_length = 0;
-        sequence_length = 0;
-        quality_length = 0;
-        is_valid = false;
+        total_bytes += storage_ptr;
+        storage_ptr = 0;
+        data_ptr = 0;
+        comm_ptr = 0;
+        qual_ptr = 0;
     };
 
-    while (!is_end) {
+    while (true) {
 
-        std::uint64_t read_bytes = gzfread(this->buffer_.data(), sizeof(char),
-            this->buffer_.size(), input_file);
-        is_end = gzeof(input_file);
-
-        total_bytes += read_bytes;
-        if (max_bytes != 0 && total_bytes > max_bytes) {
-            if (last_object_id == num_objects) {
-                throw std::invalid_argument("[bioparser::FastqParser] error: "
-                    "too small chunk size!");
+        std::uint32_t begin_ptr = this->buffer_ptr_;
+        for (; this->buffer_ptr_ < this->buffer_bytes_; ++this->buffer_ptr_) {
+            auto c = this->buffer_[this->buffer_ptr_];
+            if (c == '\n') {
+                std::memcpy(&(this->storage_[storage_ptr]),
+                    &(this->buffer_)[begin_ptr],
+                    this->buffer_ptr_ - begin_ptr);
+                storage_ptr += this->buffer_ptr_ - begin_ptr;
+                begin_ptr = this->buffer_ptr_ + 1;
+                if (line_num == 0) {
+                    data_ptr = storage_ptr;
+                    line_num = 1;
+                } else if (line_num == 2) {
+                    qual_ptr = storage_ptr;
+                    line_num = 3;
+                } else if (line_num == 3 && storage_ptr - qual_ptr == comm_ptr - data_ptr) {
+                    line_num = 0;
+                    create_T();
+                    if (total_bytes >= max_bytes) {
+                        ++this->buffer_ptr_;
+                        return true;
+                    }
+                }
+            } else if (line_num == 1 && c == '+') {
+                comm_ptr = storage_ptr;
+                line_num = 2;
             }
-            gzseek(input_file, -(current_bytes + read_bytes), SEEK_CUR);
-            status = true;
+        }
+        if (begin_ptr < this->buffer_ptr_) {
+            std::memcpy(&(this->storage_[storage_ptr]),
+                &(this->buffer_[begin_ptr]),
+                this->buffer_ptr_ - begin_ptr);
+            storage_ptr += this->buffer_ptr_ - begin_ptr;
+        }
+        this->buffer_ptr_ = 0;
+
+        if (is_eof) {
             break;
         }
 
-        for (std::uint32_t i = 0; i < read_bytes; ++i) {
-            auto c = this->buffer_[i];
+        this->buffer_bytes_ = gzread(input_file, this->buffer_.data(),
+            this->buffer_.size());
+        is_eof = this->buffer_bytes_ < this->buffer_.size();
 
-            if (c == '\n') {
-                if (!(line_number == 1 || (line_number == 3 && quality_length < sequence_length))) {
-                    line_number = (line_number + 1) % 4;
-                    if (line_number == 0) {
-                        is_valid = true;
-                    }
-                }
-            } else if (line_number == 1 && c == '+') {
-                line_number = 2;
-            } else {
-                switch (line_number) {
-                    case 0:
-                        if (name_length < kSSS) {
-                            if (!(name_length == 0 && isspace(c))) {
-                                name[name_length++] = c;
-                            }
-                        }
-                        break;
-                    case 1:
-                        sequence[sequence_length++] = c;
-                        if (sequence_length == kMSS) {
-                            this->storage_.resize(kSSS + 2 * kLSS, 0);
-                            name = &(this->storage_[0]);
-                            sequence = &(this->storage_[kSSS]);
-                            quality = &(this->storage_[kSSS + kLSS]);
-                        }
-                        break;
-                    case 2:
-                        // comment line starting with '+'
-                        // do nothing
-                        break;
-                    case 3:
-                        quality[quality_length++] = c;
-                        break;
-                    default:
-                        // never reaches this case
-                        break;
-                }
-            }
-
-            ++current_bytes;
-
-            if (is_valid) {
-                create_T();
-            }
-        }
-
-        if (is_end && current_bytes != 0) {
-            create_T();
+        if (storage_ptr + this->buffer_bytes_ > this->storage_.size()) {
+            this->storage_.resize(this->storage_.size() * 2);
         }
     }
 
-    return status;
+    if (storage_ptr != 0) {
+        create_T();
+    }
+
+    return false;
 }
 
 template<class T>
 inline MhapParser<T>::MhapParser(gzFile input_file)
-        : Parser<T>(input_file, kSSS) {
+        : Parser<T>(input_file, 65536) {
 }
 
 template<class T>
@@ -484,17 +447,10 @@ inline bool MhapParser<T>::parse(std::vector<std::unique_ptr<T>>& dst,
     std::uint64_t max_bytes, bool) {
 
     auto input_file = this->input_file_.get();
-    bool is_end = gzeof(input_file);
-    bool status = false;
-    std::uint64_t current_bytes = 0;
+    bool is_eof = false;
+
     std::uint64_t total_bytes = 0;
-    std::uint64_t num_objects = 0;
-    std::uint64_t last_object_id = num_objects;
-
-    const std::uint32_t kMhapObjectLength = 12;
-
-    char* line = &(this->storage_[0]);
-    std::uint32_t line_length = 0;
+    std::uint32_t storage_ptr = 0;
 
     std::uint64_t a_id = 0, b_id = 0;
     std::uint32_t a_rc = 0, a_begin = 0, a_end = 0, a_length = 0, b_rc = 0,
@@ -502,46 +458,46 @@ inline bool MhapParser<T>::parse(std::vector<std::unique_ptr<T>>& dst,
     double error = 0;
 
     auto create_T = [&] () -> void {
-        line[line_length] = 0;
-        rightStrip(line, line_length);
+        this->storage_[storage_ptr] = 0;
+        rightStrip(&(this->storage_[0]), storage_ptr);
 
         std::uint32_t num_values = 0, begin = 0;
         while (true) {
             std::uint32_t end = begin;
-            for (std::uint32_t j = begin; j < line_length; ++j) {
-                if (line[j] == ' ') {
+            for (std::uint32_t j = begin; j < storage_ptr; ++j) {
+                if (this->storage_[j] == ' ') {
                     end = j;
                     break;
                 }
             }
             if (end == begin) {
-                end = line_length;
+                end = storage_ptr;
             }
-            line[end] = 0;
+            this->storage_[end] = 0;
 
             switch (num_values) {
-                case 0: a_id = atoll(&line[begin]); break;
-                case 1: b_id = atoll(&line[begin]); break;
-                case 2: error = atof(&line[begin]); break;
-                case 3: minmers = atoi(&line[begin]); break;
-                case 4: a_rc = atoi(&line[begin]); break;
-                case 5: a_begin = atoi(&line[begin]); break;
-                case 6: a_end = atoi(&line[begin]); break;
-                case 7: a_length = atoi(&line[begin]); break;
-                case 8: b_rc = atoi(&line[begin]); break;
-                case 9: b_begin = atoi(&line[begin]); break;
-                case 10: b_end = atoi(&line[begin]); break;
-                case 11: b_length = atoi(&line[begin]); break;
+                case 0: a_id = atoll(&(this->storage_[begin])); break;
+                case 1: b_id = atoll(&(this->storage_[begin])); break;
+                case 2: error = atof(&(this->storage_[begin])); break;
+                case 3: minmers = atoi(&(this->storage_[begin])); break;
+                case 4: a_rc = atoi(&(this->storage_[begin])); break;
+                case 5: a_begin = atoi(&(this->storage_[begin])); break;
+                case 6: a_end = atoi(&(this->storage_[begin])); break;
+                case 7: a_length = atoi(&(this->storage_[begin])); break;
+                case 8: b_rc = atoi(&(this->storage_[begin])); break;
+                case 9: b_begin = atoi(&(this->storage_[begin])); break;
+                case 10: b_end = atoi(&(this->storage_[begin])); break;
+                case 11: b_length = atoi(&(this->storage_[begin])); break;
                 default: break;
             }
             num_values++;
-            if (end == line_length || num_values == kMhapObjectLength) {
+            if (end == storage_ptr || num_values == 12) {
                 break;
             }
             begin = end + 1;
         }
 
-        if (num_values != kMhapObjectLength) {
+        if (num_values != 12) {
             throw std::invalid_argument("[bioparser::MhapParser] error: "
                 "invalid file format!");
         }
@@ -550,51 +506,59 @@ inline bool MhapParser<T>::parse(std::vector<std::unique_ptr<T>>& dst,
             minmers, a_rc, a_begin, a_end, a_length, b_rc, b_begin,
             b_end, b_length)));
 
-        ++num_objects;
-        current_bytes = 0;
-        line_length = 0;
+        total_bytes += storage_ptr;
+        storage_ptr = 0;
     };
 
-    while (!is_end) {
+    while (true) {
 
-        std::uint64_t read_bytes = gzfread(this->buffer_.data(), sizeof(char),
-            this->buffer_.size(), input_file);
-        is_end = gzeof(input_file);
-
-        total_bytes += read_bytes;
-        if (max_bytes != 0 && total_bytes > max_bytes) {
-            if (last_object_id == num_objects) {
-                throw std::invalid_argument("[bioparser::MhapParser] error: "
-                    "too small chunk size!");
+        std::uint32_t begin_ptr = this->buffer_ptr_;
+        for (; this->buffer_ptr_ < this->buffer_bytes_; ++this->buffer_ptr_) {
+            auto c = this->buffer_[this->buffer_ptr_];
+            if (c == '\n') {
+                std::memcpy(&this->storage_[storage_ptr],
+                    &this->buffer_[begin_ptr],
+                    this->buffer_ptr_ - begin_ptr);
+                storage_ptr += this->buffer_ptr_ - begin_ptr;
+                begin_ptr = this->buffer_ptr_ + 1;
+                create_T();
+                if (total_bytes >= max_bytes) {
+                    ++this->buffer_ptr_;
+                    return true;
+                }
             }
-            gzseek(input_file, -(current_bytes + read_bytes), SEEK_CUR);
-            status = true;
+        }
+        if (begin_ptr < this->buffer_ptr_) {
+            std::memcpy(&this->storage_[storage_ptr],
+                &this->buffer_[begin_ptr],
+                this->buffer_ptr_ - begin_ptr);
+            storage_ptr += this->buffer_ptr_ - begin_ptr;
+        }
+        this->buffer_ptr_ = 0;
+
+        if (is_eof) {
             break;
         }
 
-        for (std::uint32_t i = 0; i < read_bytes; ++i) {
+        this->buffer_bytes_ = gzread(input_file, this->buffer_.data(),
+            this->buffer_.size());
+        is_eof = this->buffer_bytes_ < this->buffer_.size();
 
-            auto c = this->buffer_[i];
-            ++current_bytes;
-
-            if (c == '\n') {
-                create_T();
-            } else {
-                line[line_length++] = c;
-            }
-        }
-
-        if (is_end && current_bytes != 0) {
-            create_T();
+        if (storage_ptr + this->buffer_bytes_ > this->storage_.size()) {
+            this->storage_.resize(this->storage_.size() * 2);
         }
     }
 
-    return status;
+    if (storage_ptr != 0) {
+        create_T();
+    }
+
+    return false;
 }
 
 template<class T>
 inline PafParser<T>::PafParser(gzFile input_file)
-        : Parser<T>(input_file, 3 * kSSS + kMSS) {
+        : Parser<T>(input_file, 65536) {
 }
 
 template<class T>
@@ -606,79 +570,68 @@ inline bool PafParser<T>::parse(std::vector<std::unique_ptr<T>>& dst,
     std::uint64_t max_bytes, bool trim) {
 
     auto input_file = this->input_file_.get();
-    bool is_end = gzeof(input_file);
-    bool status = false;
-    std::uint64_t current_bytes = 0;
+    bool is_eof = false;
+
     std::uint64_t total_bytes = 0;
-    std::uint64_t num_objects = 0;
-    std::uint64_t last_object_id = num_objects;
-
-    const std::uint32_t kPafObjectLength = 12;
-
-    char* line = &(this->storage_[0]);
-    std::uint32_t line_length = 0;
+    std::uint32_t storage_ptr = 0;
 
     const char* q_name = nullptr, * t_name = nullptr;
-
     std::uint32_t q_name_length = 0, q_length = 0, q_begin = 0, q_end = 0,
         t_name_length = 0, t_length = 0, t_begin = 0, t_end = 0,
         matching_bases = 0, overlap_length = 0, mapping_quality = 0;
     char orientation = '\0';
 
     auto create_T = [&] () -> void {
-        line[line_length] = 0;
-        rightStrip(line, line_length);
+        this->storage_[storage_ptr] = 0;
+        rightStrip(&(this->storage_[0]), storage_ptr);
 
         std::uint32_t num_values = 0, begin = 0;
         while (true) {
             std::uint32_t end = begin;
-            for (std::uint32_t j = begin; j < line_length; ++j) {
-                if (line[j] == '\t') {
+            for (std::uint32_t j = begin; j < storage_ptr; ++j) {
+                if (this->storage_[j] == '\t') {
                     end = j;
                     break;
                 }
             }
             if (end == begin) {
-                end = line_length;
+                end = storage_ptr;
             }
-            line[end] = 0;
+            this->storage_[end] = 0;
 
             switch (num_values) {
                 case 0:
-                    q_name = &line[begin];
+                    q_name = &(this->storage_[begin]);
                     q_name_length = end - begin;
                     break;
-                case 1: q_length = atoi(&line[begin]); break;
-                case 2: q_begin = atoi(&line[begin]); break;
-                case 3: q_end = atoi(&line[begin]); break;
-                case 4: orientation = line[begin]; break;
+                case 1: q_length = atoi(&(this->storage_[begin])); break;
+                case 2: q_begin = atoi(&(this->storage_[begin])); break;
+                case 3: q_end = atoi(&(this->storage_[begin])); break;
+                case 4: orientation = this->storage_[begin]; break;
                 case 5:
-                    t_name = &line[begin];
+                    t_name = &(this->storage_[begin]);
                     t_name_length = end - begin;
                     break;
-                case 6: t_length = atoi(&line[begin]); break;
-                case 7: t_begin = atoi(&line[begin]); break;
-                case 8: t_end = atoi(&line[begin]); break;
-                case 9: matching_bases = atoi(&line[begin]); break;
-                case 10: overlap_length = atoi(&line[begin]); break;
-                case 11: mapping_quality = atoi(&line[begin]); break;
+                case 6: t_length = atoi(&(this->storage_[begin])); break;
+                case 7: t_begin = atoi(&(this->storage_[begin])); break;
+                case 8: t_end = atoi(&(this->storage_[begin])); break;
+                case 9: matching_bases = atoi(&(this->storage_[begin])); break;
+                case 10: overlap_length = atoi(&(this->storage_[begin])); break;
+                case 11: mapping_quality = atoi(&(this->storage_[begin])); break;
                 default: break;
             }
             num_values++;
-            if (end == line_length || num_values == kPafObjectLength) {
+            if (end == storage_ptr || num_values == 12) {
                 break;
             }
             begin = end + 1;
         }
 
-        if (num_values != kPafObjectLength) {
+        if (num_values != 12) {
             throw std::invalid_argument("[bioparser::PafParser] error: "
                 "invalid file format!");
         }
 
-        q_name_length = std::min(q_name_length, kSSS);
-        t_name_length = std::min(t_name_length, kSSS);
-
         if (trim) {
             rightStripHard(q_name, q_name_length);
             rightStripHard(t_name, t_name_length);
@@ -697,55 +650,59 @@ inline bool PafParser<T>::parse(std::vector<std::unique_ptr<T>>& dst,
             t_length, t_begin, t_end, matching_bases, overlap_length,
             mapping_quality)));
 
-        ++num_objects;
-        current_bytes = 0;
-        line_length = 0;
+        total_bytes += storage_ptr;
+        storage_ptr = 0;
     };
 
-    while (!is_end) {
-
-        std::uint64_t read_bytes = gzfread(this->buffer_.data(), sizeof(char),
-            this->buffer_.size(), input_file);
-        is_end = gzeof(input_file);
-
-        total_bytes += read_bytes;
-        if (max_bytes != 0 && total_bytes > max_bytes) {
-            if (last_object_id == num_objects) {
-                throw std::invalid_argument("[bioparser::PafParser] error: "
-                    "too small chunk size!");
-            }
-            gzseek(input_file, -(current_bytes + read_bytes), SEEK_CUR);
-            status = true;
-            break;
-        }
-
-        for (std::uint32_t i = 0; i < read_bytes; ++i) {
-
-            auto c = this->buffer_[i];
-            ++current_bytes;
+    while (true) {
 
+        std::uint32_t begin_ptr = this->buffer_ptr_;
+        for (; this->buffer_ptr_ < this->buffer_bytes_; ++this->buffer_ptr_) {
+            auto c = this->buffer_[this->buffer_ptr_];
             if (c == '\n') {
+                std::memcpy(&this->storage_[storage_ptr],
+                    &this->buffer_[begin_ptr],
+                    this->buffer_ptr_ - begin_ptr);
+                storage_ptr += this->buffer_ptr_ - begin_ptr;
+                begin_ptr = this->buffer_ptr_ + 1;
                 create_T();
-            } else {
-                line[line_length++] = c;
-                if (line_length == this->storage_.size()) {
-                    this->storage_.resize(3 * kSSS + kLSS);
-                    line = &(this->storage_[0]);
+                if (total_bytes >= max_bytes) {
+                    ++this->buffer_ptr_;
+                    return true;
                 }
             }
         }
+        if (begin_ptr < this->buffer_ptr_) {
+            std::memcpy(&this->storage_[storage_ptr],
+                &this->buffer_[begin_ptr],
+                this->buffer_ptr_ - begin_ptr);
+            storage_ptr += this->buffer_ptr_ - begin_ptr;
+        }
+        this->buffer_ptr_ = 0;
 
-        if (is_end && current_bytes != 0) {
-            create_T();
+        if (is_eof) {
+            break;
+        }
+
+        this->buffer_bytes_ = gzread(input_file, this->buffer_.data(),
+            this->buffer_.size());
+        is_eof = this->buffer_bytes_ < this->buffer_.size();
+
+        if (storage_ptr + this->buffer_bytes_ > this->storage_.size()) {
+            this->storage_.resize(this->storage_.size() * 2);
         }
     }
 
-    return status;
+    if (storage_ptr != 0) {
+        create_T();
+    }
+
+    return false;
 }
 
 template<class T>
 inline SamParser<T>::SamParser(gzFile input_file)
-        : Parser<T>(input_file, 5 * kSSS + 2 * kMSS) {
+        : Parser<T>(input_file, 65536) {
 }
 
 template<class T>
@@ -757,93 +714,85 @@ inline bool SamParser<T>::parse(std::vector<std::unique_ptr<T>>& dst,
     std::uint64_t max_bytes, bool trim) {
 
     auto input_file = this->input_file_.get();
-    bool is_end = gzeof(input_file);
-    bool status = false;
-    std::uint64_t current_bytes = 0;
+    bool is_eof = false;
+
     std::uint64_t total_bytes = 0;
-    std::uint64_t num_objects = 0;
-    std::uint64_t last_object_id = num_objects;
-
-    const std::uint32_t kSamObjectLength = 11;
-
-    char* line = &(this->storage_[0]);
-    std::uint32_t line_length = 0;
+    std::uint32_t storage_ptr = 0;
 
     const char* q_name = nullptr, * t_name = nullptr, * cigar = nullptr,
         * t_next_name = nullptr, * sequence = nullptr, * quality = nullptr;
-
     std::uint32_t q_name_length = 0, flag = 0, t_name_length = 0, t_begin = 0,
         mapping_quality = 0, cigar_length = 0, t_next_name_length = 0,
         t_next_begin = 0, template_length = 0, sequence_length = 0,
         quality_length = 0;
 
     auto create_T = [&] () -> void {
+        if (this->storage_[0] == '@') { // header
+            storage_ptr = 0;
+            return;
+        }
 
-        line[line_length] = 0;
-        rightStrip(line, line_length);
+        this->storage_[storage_ptr] = 0;
+        rightStrip(&(this->storage_[0]), storage_ptr);
 
         std::uint32_t num_values = 0, begin = 0;
         while (true) {
             std::uint32_t end = begin;
-            for (std::uint32_t j = begin; j < line_length; ++j) {
-                if (line[j] == '\t') {
+            for (std::uint32_t j = begin; j < storage_ptr; ++j) {
+                if (this->storage_[j] == '\t') {
                     end = j;
                     break;
                 }
             }
             if (end == begin) {
-                end = line_length;
+                end = storage_ptr;
             }
-            line[end] = 0;
+            this->storage_[end] = 0;
 
             switch (num_values) {
                 case 0:
-                    q_name = &line[begin];
+                    q_name = &(this->storage_[begin]);
                     q_name_length = end - begin;
                     break;
-                case 1: flag = atoi(&line[begin]); break;
+                case 1: flag = atoi(&(this->storage_[begin])); break;
                 case 2:
-                    t_name = &line[begin];
+                    t_name = &(this->storage_[begin]);
                     t_name_length = end - begin;
                     break;
-                case 3: t_begin = atoi(&line[begin]); break;
-                case 4: mapping_quality = atoi(&line[begin]); break;
+                case 3: t_begin = atoi(&(this->storage_[begin])); break;
+                case 4: mapping_quality = atoi(&(this->storage_[begin])); break;
                 case 5:
-                    cigar = &line[begin];
+                    cigar = &(this->storage_[begin]);
                     cigar_length = end - begin;
                     break;
                 case 6:
-                    t_next_name = &line[begin];
+                    t_next_name = &(this->storage_[begin]);
                     t_next_name_length = end - begin;
                     break;
-                case 7: t_next_begin = atoi(&line[begin]); break;
-                case 8: template_length = atoi(&line[begin]); break;
+                case 7: t_next_begin = atoi(&(this->storage_[begin])); break;
+                case 8: template_length = atoi(&(this->storage_[begin])); break;
                 case 9:
-                    sequence = &line[begin];
+                    sequence = &(this->storage_[begin]);
                     sequence_length = end - begin;
                     break;
                 case 10:
-                    quality = &line[begin];
+                    quality = &(this->storage_[begin]);
                     quality_length = end - begin;
                     break;
                 default: break;
             }
             num_values++;
-            if (end == line_length || num_values == kSamObjectLength) {
+            if (end == storage_ptr || num_values == 11) {
                 break;
             }
             begin = end + 1;
         }
 
-        if (num_values != kSamObjectLength) {
+        if (num_values != 11) {
             throw std::invalid_argument("[bioparser::SamParser] error: "
                 "invalid file format!");
         }
 
-        q_name_length = std::min(q_name_length, kSSS);
-        t_name_length = std::min(t_name_length, kSSS);
-        t_next_name_length = std::min(t_next_name_length, kSSS);
-
         if (trim) {
             rightStripHard(q_name, q_name_length);
             rightStripHard(t_name, t_name_length);
@@ -874,56 +823,54 @@ inline bool SamParser<T>::parse(std::vector<std::unique_ptr<T>>& dst,
             t_next_begin, template_length, sequence, sequence_length,
             quality, quality_length)));
 
-        ++num_objects;
-        current_bytes = 0;
-        line_length = 0;
+        total_bytes += storage_ptr;
+        storage_ptr = 0;
     };
 
-    while (!is_end) {
+    while (true) {
 
-        std::uint64_t read_bytes = gzfread(this->buffer_.data(), sizeof(char),
-            this->buffer_.size(), input_file);
-        is_end = gzeof(input_file);
-
-        total_bytes += read_bytes;
-        if (max_bytes != 0 && total_bytes > max_bytes) {
-            if (last_object_id == num_objects) {
-                throw std::invalid_argument("[bioparser::SamParser] error: "
-                    "too small chunk size!");
+        std::uint32_t begin_ptr = this->buffer_ptr_;
+        for (; this->buffer_ptr_ < this->buffer_bytes_; ++this->buffer_ptr_) {
+            auto c = this->buffer_[this->buffer_ptr_];
+            if (c == '\n') {
+                std::memcpy(&this->storage_[storage_ptr],
+                    &this->buffer_[begin_ptr],
+                    this->buffer_ptr_ - begin_ptr);
+                storage_ptr += this->buffer_ptr_ - begin_ptr;
+                begin_ptr = this->buffer_ptr_ + 1;
+                create_T();
+                if (total_bytes >= max_bytes) {
+                    ++this->buffer_ptr_;
+                    return true;
+                }
             }
-            gzseek(input_file, -(current_bytes + read_bytes), SEEK_CUR);
-            status = true;
+        }
+        if (begin_ptr < this->buffer_ptr_) {
+            std::memcpy(&this->storage_[storage_ptr],
+                &this->buffer_[begin_ptr],
+                this->buffer_ptr_ - begin_ptr);
+            storage_ptr += this->buffer_ptr_ - begin_ptr;
+        }
+        this->buffer_ptr_ = 0;
+
+        if (is_eof) {
             break;
         }
 
-        for (std::uint32_t i = 0; i < read_bytes; ++i) {
+        this->buffer_bytes_ = gzread(input_file, this->buffer_.data(),
+            this->buffer_.size());
+        is_eof = this->buffer_bytes_ < this->buffer_.size();
 
-            auto c = this->buffer_[i];
-            ++current_bytes;
-
-            if (c == '\n') {
-                if (line[0] == '@') {
-                    line_length = 0;
-                    current_bytes = 0;
-                    continue;
-                }
-                create_T();
-            } else {
-                line[line_length++] = c;
-                if (line_length == this->storage_.size()) {
-                    this->storage_.resize(5 * kSSS + 2 * kLSS);
-                    line = &(this->storage_[0]);
-                }
-            }
-        }
-
-        if (is_end && current_bytes != 0) {
-            create_T();
+        if (storage_ptr + this->buffer_bytes_ > this->storage_.size()) {
+            this->storage_.resize(this->storage_.size() * 2);
         }
     }
 
-    return status;
+    if (storage_ptr != 0) {
+        create_T();
+    }
+
+    return false;
 }
 
-
 }
