@@ -344,10 +344,8 @@ TEST_F(BioparserFastaTest, ParseInChunks) {
 
     SetUp(bioparser_test_data_path + "sample.fasta");
 
-    std::uint32_t size_in_bytes = 64 * 1024;
     std::vector<std::unique_ptr<Read>> reads;
-    while (parser->parse(reads, size_in_bytes)) {
-    }
+    while (parser->parse(reads, 65536));
 
     std::uint32_t name_size = 0, sequence_size = 0, quality_size = 0;
     reads_summary(name_size, sequence_size, quality_size, reads);
@@ -362,10 +360,8 @@ TEST_F(BioparserFastaTest, CompressedParseInChunks) {
 
     SetUp(bioparser_test_data_path + "sample.fasta.gz");
 
-    std::uint32_t size_in_bytes = 64 * 1024;
     std::vector<std::unique_ptr<Read>> reads;
-    while (parser->parse(reads, size_in_bytes)) {
-    }
+    while (parser->parse(reads, 65536));
 
     std::uint32_t name_size = 0, sequence_size = 0, quality_size = 0;
     reads_summary(name_size, sequence_size, quality_size, reads);
@@ -402,36 +398,6 @@ TEST_F(BioparserFastaTest, CompressedFormatError) {
     }
 }
 
-TEST_F(BioparserFastaTest, ChunkSizeError) {
-
-    SetUp(bioparser_test_data_path + "sample.fasta");
-
-    std::uint32_t size_in_bytes = 10 * 1024;
-    std::vector<std::unique_ptr<Read>> reads;
-
-    try {
-        parser->parse(reads, size_in_bytes);
-    } catch (std::invalid_argument& exception) {
-        EXPECT_STREQ(exception.what(), "[bioparser::FastaParser] error: "
-            "too small chunk size!");
-    }
-}
-
-TEST_F(BioparserFastaTest, CompressedChunkSizeError) {
-
-    SetUp(bioparser_test_data_path + "sample.fasta.gz");
-
-    std::uint32_t size_in_bytes = 10 * 1024;
-    std::vector<std::unique_ptr<Read>> reads;
-
-    try {
-        parser->parse(reads, size_in_bytes);
-    } catch (std::invalid_argument& exception) {
-        EXPECT_STREQ(exception.what(), "[bioparser::FastaParser] error: "
-            "too small chunk size!");
-    }
-}
-
 TEST_F(BioparserFastaTest, ParseAndReset) {
 
     SetUp(bioparser_test_data_path + "sample.fasta");
@@ -443,11 +409,9 @@ TEST_F(BioparserFastaTest, ParseAndReset) {
         quality_size = 0;
     reads_summary(name_size, sequence_size, quality_size, reads);
 
-    std::uint32_t size_in_bytes = 64 * 1024;
     reads.clear();
     parser->reset();
-    while (parser->parse(reads, size_in_bytes)) {
-    }
+    while (parser->parse(reads, 65536));
 
     std::uint32_t num_reads_new = reads.size(), name_size_new = 0,
         sequence_size_new = 0, quality_size_new = 0;
@@ -470,11 +434,9 @@ TEST_F(BioparserFastaTest, CompressedParseAndReset) {
         quality_size = 0;
     reads_summary(name_size, sequence_size, quality_size, reads);
 
-    std::uint32_t size_in_bytes = 64 * 1024;
     reads.clear();
     parser->reset();
-    while (parser->parse(reads, size_in_bytes)) {
-    }
+    while (parser->parse(reads, 65536));
 
     std::uint32_t num_reads_new = reads.size(), name_size_new = 0,
         sequence_size_new = 0, quality_size_new = 0;
@@ -522,10 +484,8 @@ TEST_F(BioparserFastqTest, ParseInChunks) {
 
     SetUp(bioparser_test_data_path + "sample.fastq");
 
-    std::uint32_t size_in_bytes = 64 * 1024;
     std::vector<std::unique_ptr<Read>> reads;
-    while (parser->parse(reads, size_in_bytes)) {
-    }
+    while (parser->parse(reads, 65536));
 
     std::uint32_t name_size = 0, sequence_size = 0, quality_size = 0;
     reads_summary(name_size, sequence_size, quality_size, reads);
@@ -540,10 +500,8 @@ TEST_F(BioparserFastqTest, CompressedParseInChunks) {
 
     SetUp(bioparser_test_data_path + "sample.fastq.gz");
 
-    std::uint32_t size_in_bytes = 64 * 1024;
     std::vector<std::unique_ptr<Read>> reads;
-    while (parser->parse(reads, size_in_bytes)) {
-    }
+    while (parser->parse(reads, 65536));
 
     std::uint32_t name_size = 0, sequence_size = 0, quality_size = 0;
     reads_summary(name_size, sequence_size, quality_size, reads);
@@ -582,36 +540,6 @@ TEST_F(BioparserFastqTest, CompressedFormatError) {
     }
 }
 
-TEST_F(BioparserFastqTest, ChunkSizeError) {
-
-    SetUp(bioparser_test_data_path + "sample.fastq");
-
-    std::uint32_t size_in_bytes = 10 * 1024;
-    std::vector<std::unique_ptr<Read>> reads;
-
-    try {
-        parser->parse(reads, size_in_bytes);
-    } catch (std::invalid_argument& exception) {
-        EXPECT_STREQ(exception.what(), "[bioparser::FastqParser] error: "
-            "too small chunk size!");
-    }
-}
-
-TEST_F(BioparserFastqTest, CompressedChunkSizeError) {
-
-    SetUp(bioparser_test_data_path + "sample.fastq.gz");
-
-    std::uint32_t size_in_bytes = 10 * 1024;
-    std::vector<std::unique_ptr<Read>> reads;
-
-    try {
-        parser->parse(reads, size_in_bytes);
-    } catch (std::invalid_argument& exception) {
-        EXPECT_STREQ(exception.what(), "[bioparser::FastqParser] error: "
-            "too small chunk size!");
-    }
-}
-
 TEST_F(BioparserMhapTest, ParseWhole) {
 
     SetUp(bioparser_test_data_path + "sample.mhap");
@@ -646,10 +574,8 @@ TEST_F(BioparserMhapTest, ParseInChunks) {
 
     SetUp(bioparser_test_data_path + "sample.mhap");
 
-    std::uint32_t size_in_bytes = 64 * 1024;
     std::vector<std::unique_ptr<Overlap>> overlaps;
-    while (parser->parse(overlaps, size_in_bytes)) {
-    }
+    while (parser->parse(overlaps, 1024));
 
     std::uint32_t name_size = 0, total_value = 0;
     overlaps_summary(name_size, total_value, overlaps);
@@ -663,10 +589,8 @@ TEST_F(BioparserMhapTest, CompressedParseInChunks) {
 
     SetUp(bioparser_test_data_path + "sample.mhap.gz");
 
-    std::uint32_t size_in_bytes = 64 * 1024;
     std::vector<std::unique_ptr<Overlap>> overlaps;
-    while (parser->parse(overlaps, size_in_bytes)) {
-    }
+    while (parser->parse(overlaps, 1024));
 
     std::uint32_t name_size = 0, total_value = 0;
     overlaps_summary(name_size, total_value, overlaps);
@@ -738,10 +662,8 @@ TEST_F(BioparserPafTest, ParseInChunks) {
 
     SetUp(bioparser_test_data_path + "sample.paf");
 
-    std::uint32_t size_in_bytes = 64 * 1024;
     std::vector<std::unique_ptr<Overlap>> overlaps;
-    while (parser->parse(overlaps, size_in_bytes)) {
-    }
+    while (parser->parse(overlaps, 1024));
 
     std::uint32_t name_size = 0, total_value = 0;
     overlaps_summary(name_size, total_value, overlaps);
@@ -755,10 +677,8 @@ TEST_F(BioparserPafTest, CompressedParseInChunks) {
 
     SetUp(bioparser_test_data_path + "sample.paf.gz");
 
-    std::uint32_t size_in_bytes = 64 * 1024;
     std::vector<std::unique_ptr<Overlap>> overlaps;
-    while (parser->parse(overlaps, size_in_bytes)) {
-    }
+    while (parser->parse(overlaps, 1024));
 
     std::uint32_t name_size = 0, total_value = 0;
     overlaps_summary(name_size, total_value, overlaps);
@@ -830,10 +750,8 @@ TEST_F(BioparserSamTest, ParseInChunks) {
 
     SetUp(bioparser_test_data_path + "sample.sam");
 
-    std::uint32_t size_in_bytes = 64 * 1024;
     std::vector<std::unique_ptr<Alignment>> alignments;
-    while (parser->parse(alignments, size_in_bytes)) {
-    }
+    while (parser->parse(alignments, 1024));
 
     std::uint32_t string_size = 0, total_value = 0;
     alignments_summary(string_size, total_value, alignments);
@@ -847,10 +765,8 @@ TEST_F(BioparserSamTest, CompressedParseInChunks) {
 
     SetUp(bioparser_test_data_path + "sample.sam.gz");
 
-    std::uint32_t size_in_bytes = 64 * 1024;
     std::vector<std::unique_ptr<Alignment>> alignments;
-    while (parser->parse(alignments, size_in_bytes)) {
-    }
+    while (parser->parse(alignments, 1024));
 
     std::uint32_t string_size = 0, total_value = 0;
     alignments_summary(string_size, total_value, alignments);
